# -*- coding: utf-8 -*-
#
import os
import os.path
import socket
import unittest
from base64 import decodebytes as base64decode
from unittest import mock

import websocket as ws
from websocket._abnf import ABNF
from websocket._exceptions import (
    WebSocketBadStatusException,
    WebSocketAddressException,
    WebSocketException,
)
from websocket._handshake import _create_sec_websocket_key
from websocket._handshake import _validate as _validate_header
from websocket._handshake import _get_handshake_headers
from websocket._handshake import handshake_response
from websocket._http import read_headers
from websocket._utils import validate_utf8

"""
test_websocket.py
websocket - WebSocket client library for Python

Copyright 2026 engn33r

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

try:
    import ssl
except ImportError:
    # dummy class of SSLError for ssl none-support environment.
    class SSLError(Exception):
        pass


# Skip test to access the internet unless TEST_WITH_INTERNET == 1
TEST_WITH_INTERNET = os.environ.get("TEST_WITH_INTERNET", "0") == "1"
# Skip tests relying on local websockets server unless LOCAL_WS_SERVER_PORT != -1
LOCAL_WS_SERVER_PORT = os.environ.get("LOCAL_WS_SERVER_PORT", "-1")
TEST_WITH_LOCAL_SERVER = LOCAL_WS_SERVER_PORT != "-1"
TRACEABLE = True


def create_mask_key(_):
    return "abcd"


class SockMock:
    def __init__(self):
        self.data = []
        self.sent = []

    def add_packet(self, data):
        self.data.append(data)

    def gettimeout(self):
        return None

    def recv(self, bufsize):
        if self.data:
            e = self.data.pop(0)
            if isinstance(e, Exception):
                raise e
            if len(e) > bufsize:
                self.data.insert(0, e[bufsize:])
            return e[:bufsize]

    def send(self, data):
        self.sent.append(data)
        return len(data)

    def close(self):
        pass


class HeaderSockMock(SockMock):
    def __init__(self, fname):
        SockMock.__init__(self)
        path = os.path.join(os.path.dirname(__file__), fname)
        with open(path, "rb") as f:
            self.add_packet(f.read())


class WebSocketTest(unittest.TestCase):
    def setUp(self):
        ws.enableTrace(TRACEABLE)

    def tearDown(self):
        pass

    def test_default_timeout(self):
        self.assertEqual(ws.getdefaulttimeout(), None)
        ws.setdefaulttimeout(10)
        self.assertEqual(ws.getdefaulttimeout(), 10)
        ws.setdefaulttimeout(None)

    def test_ws_key(self):
        key = _create_sec_websocket_key()
        self.assertTrue(key != 24)
        self.assertTrue("¥n" not in key)

    def test_nonce(self):
        """WebSocket key should be a random 16-byte nonce."""
        key = _create_sec_websocket_key()
        nonce = base64decode(key.encode("utf-8"))
        self.assertEqual(16, len(nonce))

    def test_ws_utils(self):
        key = "c6b8hTg4EeGb2gQMztV1/g=="
        required_header = {
            "upgrade": "websocket",
            "connection": "upgrade",
            "sec-websocket-accept": "Kxep+hNu9n51529fGidYu7a3wO0=",
        }
        self.assertEqual(_validate_header(required_header, key, None), (True, None))

        header = required_header.copy()
        header["upgrade"] = "http"
        self.assertEqual(_validate_header(header, key, None), (False, None))
        del header["upgrade"]
        self.assertEqual(_validate_header(header, key, None), (False, None))

        header = required_header.copy()
        header["connection"] = "something"
        self.assertEqual(_validate_header(header, key, None), (False, None))
        del header["connection"]
        self.assertEqual(_validate_header(header, key, None), (False, None))

        header = required_header.copy()
        header["sec-websocket-accept"] = "something"
        self.assertEqual(_validate_header(header, key, None), (False, None))
        del header["sec-websocket-accept"]
        self.assertEqual(_validate_header(header, key, None), (False, None))

        header = required_header.copy()
        header["sec-websocket-protocol"] = "sub1"
        self.assertEqual(
            _validate_header(header, key, ["sub1", "sub2"]), (True, "sub1")
        )
        # This case will print out a logging error using the error() function, but that is expected
        self.assertEqual(_validate_header(header, key, ["sub2", "sub3"]), (False, None))

        header = required_header.copy()
        header["sec-websocket-protocol"] = "sUb1"
        self.assertEqual(
            _validate_header(header, key, ["Sub1", "suB2"]), (True, "sub1")
        )

        header = required_header.copy()
        # This case will print out a logging error using the error() function, but that is expected
        self.assertEqual(_validate_header(header, key, ["Sub1", "suB2"]), (False, None))

    def test_read_header(self):
        status, header, _ = read_headers(HeaderSockMock("data/header01.txt"))
        self.assertEqual(status, 101)
        self.assertEqual(header["connection"], "Upgrade")

        status, header, _ = read_headers(HeaderSockMock("data/header03.txt"))
        self.assertEqual(status, 101)
        self.assertEqual(header["connection"], "Upgrade, Keep-Alive")

        HeaderSockMock("data/header02.txt")
        self.assertRaises(
            ws.WebSocketException, read_headers, HeaderSockMock("data/header02.txt")
        )

    def test_send(self):
        # TODO: add longer frame data
        sock = ws.WebSocket()
        sock.set_mask_key(create_mask_key)
        s = sock.sock = HeaderSockMock("data/header01.txt")
        sock.send("Hello")
        self.assertEqual(s.sent[0], b"\x81\x85abcd)\x07\x0f\x08\x0e")

        sock.send("こんにちは")
        self.assertEqual(
            s.sent[1],
            b"\x81\x8fabcd\x82\xe3\xf0\x87\xe3\xf1\x80\xe5\xca\x81\xe2\xc5\x82\xe3\xcc",
        )

        #        sock.send("x" * 5000)
        #        self.assertEqual(s.sent[1], b'\x81\x8fabcd\x82\xe3\xf0\x87\xe3\xf1\x80\xe5\xca\x81\xe2\xc5\x82\xe3\xcc")

        self.assertEqual(sock.send_binary(b"1111111111101"), 19)

    def test_send_close_accepts_text_reason(self):
        sock = ws.WebSocket()
        captured = {}

        def fake_send(payload, opcode):
            captured["payload"] = payload
            captured["opcode"] = opcode
            return len(payload)

        sock.send = fake_send  # type: ignore[assignment]
        sock.connected = True

        sock.send_close(reason="normal close")

        self.assertEqual(captured["opcode"], ws.ABNF.OPCODE_CLOSE)
        self.assertIsInstance(captured["payload"], (bytes, bytearray))
        self.assertTrue(captured["payload"].endswith(b"normal close"))

    def test_close_accepts_text_reason(self):
        sock = ws.WebSocket()
        captured = {}

        def fake_send(payload, opcode):
            captured["payload"] = payload
            captured["opcode"] = opcode
            return len(payload)

        sock.send = fake_send  # type: ignore[assignment]
        sock.connected = True

        sock.close(reason="normal close")

        self.assertEqual(captured["opcode"], ws.ABNF.OPCODE_CLOSE)
        self.assertIsInstance(captured["payload"], (bytes, bytearray))
        self.assertTrue(captured["payload"].endswith(b"normal close"))

    def test_recv(self):
        # TODO: add longer frame data
        sock = ws.WebSocket()
        s = sock.sock = SockMock()
        something = (
            b"\x81\x8fabcd\x82\xe3\xf0\x87\xe3\xf1\x80\xe5\xca\x81\xe2\xc5\x82\xe3\xcc"
        )
        s.add_packet(something)
        data = sock.recv()
        self.assertEqual(data, "こんにちは")

        s.add_packet(b"\x81\x85abcd)\x07\x0f\x08\x0e")
        data = sock.recv()
        self.assertEqual(data, "Hello")

    @unittest.skipUnless(TEST_WITH_INTERNET, "Internet-requiring tests are disabled")
    def test_iter(self):
        count = 2
        s = ws.create_connection("wss://api.bitfinex.com/ws/2")
        try:
            s.send('{"event": "subscribe", "channel": "ticker"}')
            for _ in s:
                count -= 1
                if count == 0:
                    break
            self.assertEqual(
                count, 0, "WebSocket iterator failed to yield the expected frames"
            )
        finally:
            s.close()

    @unittest.skipUnless(TEST_WITH_INTERNET, "Internet-requiring tests are disabled")
    def test_next(self):
        sock = ws.create_connection("wss://api.bitfinex.com/ws/2")
        try:
            self.assertEqual(str, type(next(sock)))
        finally:
            sock.close()

    def test_internal_recv_strict(self):
        sock = ws.WebSocket()
        s = sock.sock = SockMock()
        s.add_packet(b"foo")
        s.add_packet(socket.timeout())
        s.add_packet(b"bar")
        # s.add_packet(SSLError("The read operation timed out"))
        s.add_packet(b"baz")
        with self.assertRaises(ws.WebSocketTimeoutException):
            sock.frame_buffer.recv_strict(9)
        #     with self.assertRaises(SSLError):
        #         data = sock._recv_strict(9)
        data = sock.frame_buffer.recv_strict(9)
        self.assertEqual(data, b"foobarbaz")
        with self.assertRaises(ws.WebSocketConnectionClosedException):
            sock.frame_buffer.recv_strict(1)

    def test_recv_timeout(self):
        sock = ws.WebSocket()
        s = sock.sock = SockMock()
        s.add_packet(b"\x81")
        s.add_packet(socket.timeout())
        s.add_packet(b"\x8dabcd\x29\x07\x0f\x08\x0e")
        s.add_packet(socket.timeout())
        s.add_packet(b"\x4e\x43\x33\x0e\x10\x0f\x00\x40")
        with self.assertRaises(ws.WebSocketTimeoutException):
            sock.recv()
        with self.assertRaises(ws.WebSocketTimeoutException):
            sock.recv()
        data = sock.recv()
        self.assertEqual(data, "Hello, World!")
        with self.assertRaises(ws.WebSocketConnectionClosedException):
            sock.recv()

    def test_recv_with_simple_fragmentation(self):
        sock = ws.WebSocket()
        s = sock.sock = SockMock()
        # OPCODE=TEXT, FIN=0, MSG="Brevity is "
        s.add_packet(b"\x01\x8babcd#\x10\x06\x12\x08\x16\x1aD\x08\x11C")
        # OPCODE=CONT, FIN=1, MSG="the soul of wit"
        s.add_packet(b"\x80\x8fabcd\x15\n\x06D\x12\r\x16\x08A\r\x05D\x16\x0b\x17")
        data = sock.recv()
        self.assertEqual(data, "Brevity is the soul of wit")
        with self.assertRaises(ws.WebSocketConnectionClosedException):
            sock.recv()

    def test_recv_with_fire_event_of_fragmentation(self):
        sock = ws.WebSocket(fire_cont_frame=True)
        s = sock.sock = SockMock()
        # OPCODE=TEXT, FIN=0, MSG="Brevity is "
        s.add_packet(b"\x01\x8babcd#\x10\x06\x12\x08\x16\x1aD\x08\x11C")
        # OPCODE=CONT, FIN=0, MSG="Brevity is "
        s.add_packet(b"\x00\x8babcd#\x10\x06\x12\x08\x16\x1aD\x08\x11C")
        # OPCODE=CONT, FIN=1, MSG="the soul of wit"
        s.add_packet(b"\x80\x8fabcd\x15\n\x06D\x12\r\x16\x08A\r\x05D\x16\x0b\x17")

        _, data = sock.recv_data()
        self.assertEqual(data, b"Brevity is ")
        _, data = sock.recv_data()
        self.assertEqual(data, b"Brevity is ")
        _, data = sock.recv_data()
        self.assertEqual(data, b"the soul of wit")

        # OPCODE=CONT, FIN=0, MSG="Brevity is "
        s.add_packet(b"\x80\x8babcd#\x10\x06\x12\x08\x16\x1aD\x08\x11C")

        with self.assertRaises(ws.WebSocketException):
            sock.recv_data()

        with self.assertRaises(ws.WebSocketConnectionClosedException):
            sock.recv()

    def test_close(self):
        sock = ws.WebSocket()
        sock.connected = True
        sock.close()

        sock = ws.WebSocket()
        s = sock.sock = SockMock()
        sock.connected = True
        s.add_packet(b"\x88\x80\x17\x98p\x84")
        sock.recv()
        self.assertEqual(sock.connected, False)

    def test_recv_cont_fragmentation(self):
        sock = ws.WebSocket()
        s = sock.sock = SockMock()
        # OPCODE=CONT, FIN=1, MSG="the soul of wit"
        s.add_packet(b"\x80\x8fabcd\x15\n\x06D\x12\r\x16\x08A\r\x05D\x16\x0b\x17")
        self.assertRaises(ws.WebSocketException, sock.recv)

    def test_recv_with_prolonged_fragmentation(self):
        sock = ws.WebSocket()
        s = sock.sock = SockMock()
        # OPCODE=TEXT, FIN=0, MSG="Once more unto the breach, "
        s.add_packet(
            b"\x01\x9babcd.\x0c\x00\x01A\x0f\x0c\x16\x04B\x16\n\x15\rC\x10\t\x07C\x06\x13\x07\x02\x07\tNC"
        )
        # OPCODE=CONT, FIN=0, MSG="dear friends, "
        s.add_packet(b"\x00\x8eabcd\x05\x07\x02\x16A\x04\x11\r\x04\x0c\x07\x17MB")
        # OPCODE=CONT, FIN=1, MSG="once more"
        s.add_packet(b"\x80\x89abcd\x0e\x0c\x00\x01A\x0f\x0c\x16\x04")
        data = sock.recv()
        self.assertEqual(data, "Once more unto the breach, dear friends, once more")
        with self.assertRaises(ws.WebSocketConnectionClosedException):
            sock.recv()

    def test_recv_with_fragmentation_and_control_frame(self):
        sock = ws.WebSocket()
        sock.set_mask_key(create_mask_key)
        s = sock.sock = SockMock()
        # OPCODE=TEXT, FIN=0, MSG="Too much "
        s.add_packet(b"\x01\x89abcd5\r\x0cD\x0c\x17\x00\x0cA")
        # OPCODE=PING, FIN=1, MSG="Please PONG this"
        s.add_packet(b"\x89\x90abcd1\x0e\x06\x05\x12\x07C4.,$D\x15\n\n\x17")
        # OPCODE=CONT, FIN=1, MSG="of a good thing"
        s.add_packet(b"\x80\x8fabcd\x0e\x04C\x05A\x05\x0c\x0b\x05B\x17\x0c\x08\x0c\x04")
        data = sock.recv()
        self.assertEqual(data, "Too much of a good thing")
        with self.assertRaises(ws.WebSocketConnectionClosedException):
            sock.recv()
        self.assertEqual(
            s.sent[0], b"\x8a\x90abcd1\x0e\x06\x05\x12\x07C4.,$D\x15\n\n\x17"
        )

    @unittest.skipUnless(
        TEST_WITH_LOCAL_SERVER, "Tests using local websocket server are disabled"
    )
    def test_websocket(self):
        s = ws.create_connection(f"ws://127.0.0.1:{LOCAL_WS_SERVER_PORT}")
        self.assertNotEqual(s, None)
        s.send("Hello, World")
        result = s.next()
        s.fileno()
        self.assertEqual(result, "Hello, World")

        s.send("こにゃにゃちは、世界")
        result = s.recv()
        self.assertEqual(result, "こにゃにゃちは、世界")
        self.assertRaises(ValueError, s.send_close, -1, "")
        s.close()

    @unittest.skipUnless(
        TEST_WITH_LOCAL_SERVER, "Tests using local websocket server are disabled"
    )
    def test_ping_pong(self):
        s = ws.create_connection(f"ws://127.0.0.1:{LOCAL_WS_SERVER_PORT}")
        self.assertNotEqual(s, None)
        s.ping("Hello")
        s.pong("Hi")
        s.close()

    @unittest.skipUnless(TEST_WITH_INTERNET, "Internet-requiring tests are disabled")
    def test_support_redirect(self):
        s = ws.WebSocket()
        self.assertRaises(WebSocketBadStatusException, s.connect, "ws://google.com/")
        # Need to find a URL that has a redirect code leading to a websocket

    @unittest.skipUnless(TEST_WITH_INTERNET, "Internet-requiring tests are disabled")
    def test_secure_websocket(self):
        s = ws.create_connection("wss://api.bitfinex.com/ws/2")
        self.assertNotEqual(s, None)
        self.assertTrue(isinstance(s.sock, ssl.SSLSocket))
        self.assertEqual(s.getstatus(), 101)
        self.assertNotEqual(s.getheaders(), None)
        s.settimeout(10)
        self.assertEqual(s.gettimeout(), 10)
        self.assertEqual(s.getsubprotocol(), None)
        s.abort()

    @unittest.skipUnless(
        TEST_WITH_LOCAL_SERVER, "Tests using local websocket server are disabled"
    )
    def test_websocket_with_custom_header(self):
        s = ws.create_connection(
            f"ws://127.0.0.1:{LOCAL_WS_SERVER_PORT}",
            header={"User-Agent": "PythonWebsocketClient"},
        )
        self.assertNotEqual(s, None)
        self.assertEqual(s.getsubprotocol(), None)
        s.send("Hello, World")
        result = s.recv()
        self.assertEqual(result, "Hello, World")
        self.assertRaises(ValueError, s.close, -1, "")
        s.close()

    @unittest.skipUnless(
        TEST_WITH_LOCAL_SERVER, "Tests using local websocket server are disabled"
    )
    def test_after_close(self):
        s = ws.create_connection(f"ws://127.0.0.1:{LOCAL_WS_SERVER_PORT}")
        self.assertNotEqual(s, None)
        s.close()
        self.assertRaises(ws.WebSocketConnectionClosedException, s.send, "Hello")
        self.assertRaises(ws.WebSocketConnectionClosedException, s.recv)


class SockOptTest(unittest.TestCase):
    @unittest.skipUnless(
        TEST_WITH_LOCAL_SERVER, "Tests using local websocket server are disabled"
    )
    def test_sockopt(self):
        sockopt = ((socket.IPPROTO_TCP, socket.TCP_NODELAY, 1),)
        s = ws.create_connection(
            f"ws://127.0.0.1:{LOCAL_WS_SERVER_PORT}", sockopt=sockopt
        )
        self.assertNotEqual(
            s.sock.getsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY), 0
        )
        s.close()


class UtilsTest(unittest.TestCase):
    def test_utf8_validator(self):
        state = validate_utf8(b"\xf0\x90\x80\x80")
        self.assertEqual(state, True)
        state = validate_utf8(
            b"\xce\xba\xe1\xbd\xb9\xcf\x83\xce\xbc\xce\xb5\xed\xa0\x80edited"
        )
        self.assertEqual(state, False)
        state = validate_utf8(b"")
        self.assertEqual(state, True)
        state = validate_utf8("κόσμε")
        self.assertEqual(state, True)
        state = validate_utf8("a\ud800b")
        self.assertEqual(state, False)
        state = validate_utf8("")
        self.assertEqual(state, True)


class HandshakeTest(unittest.TestCase):
    def test_bad_status_exception_carries_status_message(self):
        exc = WebSocketBadStatusException(
            "Handshake status 404 Not Found", 404, "Not Found"
        )
        self.assertEqual(exc.status_code, 404)
        self.assertEqual(exc.status_message, "Not Found")
        self.assertIsNone(exc.resp_headers)
        self.assertIsNone(exc.resp_body)

    @unittest.skipUnless(TEST_WITH_INTERNET, "Internet-requiring tests are disabled")
    def test_http_ssl(self):
        websock1 = ws.WebSocket(
            sslopt={"cert_chain": ssl.get_default_verify_paths().capath},
            enable_multithread=False,
        )
        self.assertRaises(ValueError, websock1.connect, "wss://api.bitfinex.com/ws/2")
        websock2 = ws.WebSocket(sslopt={"certfile": "myNonexistentCertFile"})
        self.assertRaises(
            WebSocketException, websock2.connect, "wss://api.bitfinex.com/ws/2"
        )

    @unittest.skipUnless(TEST_WITH_INTERNET, "Internet-requiring tests are disabled")
    def test_manual_headers(self):
        websock3 = ws.WebSocket(
            sslopt={
                "ca_certs": ssl.get_default_verify_paths().cafile,
                "ca_cert_path": ssl.get_default_verify_paths().capath,
            }
        )
        self.assertRaises(
            WebSocketBadStatusException,
            websock3.connect,
            "wss://api.bitfinex.com/ws/2",
            cookie="chocolate",
            origin="testing_websockets.com",
            host="echo.websocket.events/websocket-client-test",
            subprotocols=["testproto"],
            connection="Upgrade",
            header={
                "CustomHeader1": "123",
                "Cookie": "TestValue",
                "Sec-WebSocket-Key": "k9kFAUWNAMmf5OEMfTlOEA==",
                "Sec-WebSocket-Protocol": "newprotocol",
            },
        )

    def test_ipv6(self):
        websock2 = ws.WebSocket()
        self.assertRaises(ValueError, websock2.connect, "2001:4860:4860::8888")

    def test_bad_urls(self):
        websock3 = ws.WebSocket()
        self.assertRaises(ValueError, websock3.connect, "ws//example.com")
        self.assertRaises(WebSocketAddressException, websock3.connect, "ws://example")
        self.assertRaises(ValueError, websock3.connect, "example.com")

    def test_suppress_host_header(self):
        """Test suppress_host parameter in _get_handshake_headers function"""
        # Test normal behavior (Host header included)
        headers, key = _get_handshake_headers(
            "/path", "ws://example.com:8080", "example.com", 8080, {}
        )
        host_headers = [h for h in headers if h.startswith("Host:")]
        self.assertEqual(len(host_headers), 1)
        self.assertEqual(host_headers[0], "Host: example.com:8080")

        # Test suppress_host=False (explicit, should still include Host header)
        headers, key = _get_handshake_headers(
            "/path",
            "ws://example.com:8080",
            "example.com",
            8080,
            {"suppress_host": False},
        )
        host_headers = [h for h in headers if h.startswith("Host:")]
        self.assertEqual(len(host_headers), 1)
        self.assertEqual(host_headers[0], "Host: example.com:8080")

        # Test suppress_host=True (Host header should be suppressed)
        headers, key = _get_handshake_headers(
            "/path",
            "ws://example.com:8080",
            "example.com",
            8080,
            {"suppress_host": True},
        )
        host_headers = [h for h in headers if h.startswith("Host:")]
        self.assertEqual(len(host_headers), 0)

        # Test with custom host header and suppress_host=False
        headers, key = _get_handshake_headers(
            "/path",
            "ws://example.com:8080",
            "example.com",
            8080,
            {"host": "custom-host.example.com", "suppress_host": False},
        )
        host_headers = [h for h in headers if h.startswith("Host:")]
        self.assertEqual(len(host_headers), 1)
        self.assertEqual(host_headers[0], "Host: custom-host.example.com")

        # Test with custom host header and suppress_host=True (should suppress even custom host)
        headers, key = _get_handshake_headers(
            "/path",
            "ws://example.com:8080",
            "example.com",
            8080,
            {"host": "custom-host.example.com", "suppress_host": True},
        )
        host_headers = [h for h in headers if h.startswith("Host:")]
        self.assertEqual(len(host_headers), 0)

        # Test with standard ports (80, 443) - should not include port in host
        headers, key = _get_handshake_headers(
            "/path", "ws://example.com", "example.com", 80, {}
        )
        host_headers = [h for h in headers if h.startswith("Host:")]
        self.assertEqual(len(host_headers), 1)
        self.assertEqual(host_headers[0], "Host: example.com")

        # Test suppress_host=True with standard port
        headers, key = _get_handshake_headers(
            "/path", "ws://example.com", "example.com", 80, {"suppress_host": True}
        )
        host_headers = [h for h in headers if h.startswith("Host:")]
        self.assertEqual(len(host_headers), 0)

    def test_suppress_host_websocket_connect(self):
        """Test suppress_host parameter with WebSocket.connect()"""
        websock = ws.WebSocket()

        # Test that suppress_host parameter is accepted without error
        # (Connection will fail but parameter should be accepted)
        try:
            websock.connect(
                "ws://nonexistent.example.com", suppress_host=True, timeout=0.1
            )
        except (WebSocketAddressException, OSError, socket.timeout):
            # Expected - connection should fail, but parameter was accepted
            pass

        # Test that suppress_host=False also works
        try:
            websock.connect(
                "ws://nonexistent.example.com", suppress_host=False, timeout=0.1
            )
        except (WebSocketAddressException, OSError, socket.timeout):
            # Expected - connection should fail, but parameter was accepted
            pass


class WebSocketCoreUnitTests(unittest.TestCase):
    def test_iteration_helpers_and_next_aliases(self):
        class TestError(Exception):
            pass

        sock = ws.WebSocket()
        sock.recv = mock.Mock(side_effect=["first", TestError("stop")])

        iterator = iter(sock)
        self.assertEqual(next(iterator), "first")
        with self.assertRaises(TestError):
            next(iterator)

        sock.recv = mock.Mock(return_value="again")
        self.assertEqual(sock.__next__(), "again")
        self.assertEqual(sock.next(), "again")

    def test_fileno_requires_socket(self):
        sock = ws.WebSocket()
        with self.assertRaises(WebSocketException):
            sock.fileno()

        fake_socket = mock.Mock()
        fake_socket.fileno.return_value = 42
        sock.sock = fake_socket
        self.assertEqual(sock.fileno(), 42)

    def test_settimeout_updates_underlying_socket(self):
        sock = ws.WebSocket()
        fake_socket = mock.Mock()
        sock.sock = fake_socket

        sock.settimeout(5)
        self.assertEqual(sock.gettimeout(), 5)
        fake_socket.settimeout.assert_called_once_with(5)

    def test_is_ssl_detection(self):
        sock = ws.WebSocket()
        self.assertFalse(sock.is_ssl())

        class FakeSSLSocket:
            pass

        with mock.patch("websocket._core.ssl.SSLSocket", FakeSSLSocket):
            sock.sock = FakeSSLSocket()
            self.assertTrue(sock.is_ssl())

    def test_send_frame_emits_trace_when_enabled(self):
        sock = ws.WebSocket()
        sock._send = mock.Mock(side_effect=lambda data: len(data))
        sock.lock = mock.MagicMock()

        frame = ABNF.create_frame("hi", ABNF.OPCODE_TEXT)

        with mock.patch(
            "websocket._core.isEnabledForTrace", return_value=True
        ), mock.patch("websocket._core.trace") as trace_mock:
            length = sock.send_frame(frame)

        self.assertGreater(length, 0)
        self.assertEqual(trace_mock.call_count, 2)

    def test_connect_handles_redirects(self):
        sock = ws.WebSocket()
        first_socket = mock.Mock()
        second_socket = mock.Mock()
        connect_results = [
            (first_socket, ("origin", 80, "/")),
            (second_socket, ("redirect", 80, "/")),
        ]

        def fake_connect(url, *_):
            return connect_results.pop(0)

        redirect_resp = handshake_response(301, {"location": "ws://redirect"}, None)
        success_resp = handshake_response(101, {}, None)

        with mock.patch(
            "websocket._core.connect", side_effect=fake_connect
        ) as connect_mock, mock.patch(
            "websocket._core.handshake", side_effect=[redirect_resp, success_resp]
        ) as handshake_mock:
            sock.connect("ws://origin", redirect_limit=1)

        self.assertTrue(sock.connected)
        self.assertIs(sock.sock, second_socket)
        first_socket.close.assert_called_once()
        self.assertEqual(
            [call.args[0] for call in connect_mock.call_args_list],
            ["ws://origin", "ws://redirect"],
        )
        self.assertEqual(handshake_mock.call_count, 2)

    def test_connect_redirect_limit_exhausted_raises(self):
        sock = ws.WebSocket()
        sockets = []
        redirect_resp = handshake_response(302, {"location": "ws://redirect"}, None)

        def fake_connect(url, *args, **kwargs):
            socket_mock = mock.Mock()
            sockets.append(socket_mock)
            return socket_mock, ("redirect", 80, "/")

        with mock.patch(
            "websocket._core.connect", side_effect=fake_connect
        ) as connect_mock, mock.patch(
            "websocket._core.handshake", return_value=redirect_resp
        ):
            with self.assertRaisesRegex(
                ws.WebSocketException, "Redirect limit exhausted"
            ):
                sock.connect("ws://origin", redirect_limit=2)

        self.assertEqual(connect_mock.call_count, 3)
        self.assertFalse(sock.connected)
        self.assertIsNone(sock.sock)
        for socket_mock in sockets:
            socket_mock.close.assert_called_once()

    def test_connect_redirect_invalid_target_raises(self):
        sock = ws.WebSocket()
        redirect_resp = handshake_response(302, {"location": "https://redirect"}, None)

        def fake_connect(url, *args, **kwargs):
            if url == "https://redirect":
                raise ValueError("scheme https is invalid")
            return mock.Mock(), ("origin", 80, "/")

        with mock.patch(
            "websocket._core.connect", side_effect=fake_connect
        ), mock.patch("websocket._core.handshake", return_value=redirect_resp):
            with self.assertRaisesRegex(
                ws.WebSocketException, "Invalid redirect target 'https://redirect'"
            ):
                sock.connect("ws://origin", redirect_limit=2)

        self.assertFalse(sock.connected)
        self.assertIsNone(sock.sock)

    def test_connect_redirect_without_location_raises(self):
        sock = ws.WebSocket()
        redirect_resp = handshake_response(302, {}, None)  # no location header

        def fake_connect(url, *args, **kwargs):
            return mock.Mock(), ("origin", 80, "/")

        with mock.patch(
            "websocket._core.connect", side_effect=fake_connect
        ), mock.patch("websocket._core.handshake", return_value=redirect_resp):
            with self.assertRaisesRegex(ws.WebSocketException, "Location"):
                sock.connect("ws://origin", redirect_limit=2)

        self.assertFalse(sock.connected)
        self.assertIsNone(sock.sock)

    def test_shutdown_and_abort(self):
        sock = ws.WebSocket()
        socket_mock = mock.Mock()
        socket_mock._closed = False
        sock.sock = socket_mock
        sock.connected = True

        sock.shutdown()
        socket_mock.close.assert_called_once()
        self.assertIsNone(sock.sock)
        self.assertFalse(sock.connected)

        socket_mock = mock.Mock()
        sock.sock = socket_mock
        sock.connected = True
        sock.abort()
        socket_mock.shutdown.assert_called_once_with(socket.SHUT_RDWR)

    def test_abort_tolerates_dead_socket(self):
        sock = ws.WebSocket()
        # never-connected socket: shutdown() raises ENOTCONN
        # which abort() must not leak to the caller
        real = socket.socket()
        sock.sock = real
        sock.connected = True

        sock.abort()  # must not raise
        real.close()

    def test_create_connection_uses_custom_class(self):
        class DummySocket:
            def __init__(self, **kwargs):
                self.init_kwargs = kwargs
                self.timeout_value = None
                self.url = None
                self.options = None

            def settimeout(self, value):
                self.timeout_value = value

            def connect(self, url, **options):
                self.url = url
                self.options = options

        conn = ws.create_connection(
            "ws://example.com",
            timeout=5,
            class_=DummySocket,
            header={"User-Agent": "tester"},
            origin="https://origin",
        )

        self.assertIsInstance(conn, DummySocket)
        self.assertEqual(conn.url, "ws://example.com")
        self.assertEqual(conn.timeout_value, 5)
        self.assertIn("header", conn.options)
        self.assertEqual(conn.options["origin"], "https://origin")


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
#
import os
import os.path
import socket
import ssl
import struct
import threading
import unittest
from unittest import mock

import websocket as ws
from websocket._abnf import ABNF

"""
test_app.py
websocket - WebSocket client library for Python

Copyright 2026 engn33r

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# Skip test to access the internet unless TEST_WITH_INTERNET == 1
TEST_WITH_INTERNET = os.environ.get("TEST_WITH_INTERNET", "0") == "1"
# Skip tests relying on local websockets server unless LOCAL_WS_SERVER_PORT != -1
LOCAL_WS_SERVER_PORT = os.environ.get("LOCAL_WS_SERVER_PORT", "-1")
TEST_WITH_LOCAL_SERVER = LOCAL_WS_SERVER_PORT != "-1"
TRACEABLE = True


class WebSocketAppTest(unittest.TestCase):
    class NotSetYet:
        """A marker class for signalling that a value hasn't been set yet."""

    def setUp(self):
        ws.enableTrace(TRACEABLE)

        WebSocketAppTest.keep_running_open = WebSocketAppTest.NotSetYet()
        WebSocketAppTest.keep_running_close = WebSocketAppTest.NotSetYet()
        WebSocketAppTest.get_mask_key_id = WebSocketAppTest.NotSetYet()
        WebSocketAppTest.on_error_data = WebSocketAppTest.NotSetYet()

    def tearDown(self):
        WebSocketAppTest.keep_running_open = WebSocketAppTest.NotSetYet()
        WebSocketAppTest.keep_running_close = WebSocketAppTest.NotSetYet()
        WebSocketAppTest.get_mask_key_id = WebSocketAppTest.NotSetYet()
        WebSocketAppTest.on_error_data = WebSocketAppTest.NotSetYet()

    def close(self):
        pass

    @unittest.skipUnless(
        TEST_WITH_LOCAL_SERVER, "Tests using local websocket server are disabled"
    )
    def test_keep_running(self):
        """A WebSocketApp should keep running as long as its self.keep_running
        is not False (in the boolean context).
        """

        def on_open(self, *args, **kwargs):
            """Set the keep_running flag for later inspection and immediately
            close the connection.
            """
            self.send("hello!")
            WebSocketAppTest.keep_running_open = self.keep_running
            self.keep_running = False

        def on_message(app, message):
            print(message)
            app.close()

        def on_close(self, *args, **kwargs):
            """Set the keep_running flag for the test to use."""
            WebSocketAppTest.keep_running_close = self.keep_running

        app = ws.WebSocketApp(
            f"ws://127.0.0.1:{LOCAL_WS_SERVER_PORT}",
            on_open=on_open,
            on_close=on_close,
            on_message=on_message,
        )
        app.run_forever()

    #    @unittest.skipUnless(TEST_WITH_LOCAL_SERVER, "Tests using local websocket server are disabled")
    @unittest.skipUnless(False, "Test disabled for now (requires rel)")
    def test_run_forever_dispatcher(self):
        """A WebSocketApp should keep running as long as its self.keep_running
        is not False (in the boolean context).
        """

        def on_open(self, *args, **kwargs):
            """Send a message, receive, and send one more"""
            self.send("hello!")
            if self.sock:
                self.sock.recv()
            self.send("goodbye!")

        def on_message(app, message):
            print(message)
            app.close()

        app = ws.WebSocketApp(
            f"ws://127.0.0.1:{LOCAL_WS_SERVER_PORT}",
            on_open=on_open,
            on_message=on_message,
        )
        app.run_forever(dispatcher="Dispatcher")  # doesn't work

    #        app.run_forever(dispatcher=rel)          # would work
    #        rel.dispatch()

    @unittest.skipUnless(
        TEST_WITH_LOCAL_SERVER, "Tests using local websocket server are disabled"
    )
    def test_run_forever_teardown_clean_exit(self):
        """The WebSocketApp.run_forever() method should return `False` when the application ends gracefully."""
        app = ws.WebSocketApp(f"ws://127.0.0.1:{LOCAL_WS_SERVER_PORT}")
        threading.Timer(interval=0.2, function=app.close).start()
        teardown = app.run_forever()
        self.assertEqual(teardown, False)

    @unittest.skipUnless(TEST_WITH_INTERNET, "Internet-requiring tests are disabled")
    def test_sock_mask_key(self):
        """A WebSocketApp should forward the received mask_key function down
        to the actual socket.
        """

        def my_mask_key_func():
            return "\x00\x00\x00\x00"

        app = ws.WebSocketApp(
            "wss://api-pub.bitfinex.com/ws/1", get_mask_key=my_mask_key_func
        )

        # if numpy is installed, this assertion fail
        # Note: We can't use 'is' for comparing the functions directly, need to use 'id'.
        self.assertEqual(id(app.get_mask_key), id(my_mask_key_func))

    @unittest.skipUnless(TEST_WITH_INTERNET, "Internet-requiring tests are disabled")
    def test_invalid_ping_interval_ping_timeout(self):
        """Test exception handling if ping_interval < ping_timeout"""

        def on_ping(app, _):
            print("Got a ping!")
            app.close()

        def on_pong(app, _):
            print("Got a pong! No need to respond")
            app.close()

        app = ws.WebSocketApp(
            "wss://api-pub.bitfinex.com/ws/1", on_ping=on_ping, on_pong=on_pong
        )
        self.assertRaises(
            ws.WebSocketException,
            app.run_forever,
            ping_interval=1,
            ping_timeout=2,
            sslopt={"cert_reqs": ssl.CERT_NONE},
        )

    @unittest.skipUnless(TEST_WITH_INTERNET, "Internet-requiring tests are disabled")
    def test_ping_interval(self):
        """Test WebSocketApp proper ping functionality"""

        def on_ping(app, _):
            print("Got a ping!")
            app.close()

        def on_pong(app, _):
            print("Got a pong! No need to respond")
            app.close()

        app = ws.WebSocketApp(
            "wss://api-pub.bitfinex.com/ws/1", on_ping=on_ping, on_pong=on_pong
        )
        app.run_forever(
            ping_interval=2, ping_timeout=1, sslopt={"cert_reqs": ssl.CERT_NONE}
        )

    @unittest.skipUnless(TEST_WITH_INTERNET, "Internet-requiring tests are disabled")
    def test_opcode_close(self):
        """Test WebSocketApp close opcode"""

        app = ws.WebSocketApp("wss://tsock.us1.twilio.com/v3/wsconnect")
        app.run_forever(ping_interval=2, ping_timeout=1, ping_payload="Ping payload")

    # This is commented out because the URL no longer responds in the expected way
    # @unittest.skipUnless(TEST_WITH_INTERNET, "Internet-requiring tests are disabled")
    # def testOpcodeBinary(self):
    #     """ Test WebSocketApp binary opcode
    #     """
    #     app = ws.WebSocketApp('wss://streaming.vn.teslamotors.com/streaming/')
    #     app.run_forever(ping_interval=2, ping_timeout=1, ping_payload="Ping payload")

    @unittest.skipUnless(TEST_WITH_INTERNET, "Internet-requiring tests are disabled")
    def test_bad_ping_interval(self):
        """A WebSocketApp handling of negative ping_interval"""
        app = ws.WebSocketApp("wss://api-pub.bitfinex.com/ws/1")
        self.assertRaises(
            ws.WebSocketException,
            app.run_forever,
            ping_interval=-5,
            sslopt={"cert_reqs": ssl.CERT_NONE},
        )

    @unittest.skipUnless(TEST_WITH_INTERNET, "Internet-requiring tests are disabled")
    def test_bad_ping_timeout(self):
        """A WebSocketApp handling of negative ping_timeout"""
        app = ws.WebSocketApp("wss://api-pub.bitfinex.com/ws/1")
        self.assertRaises(
            ws.WebSocketException,
            app.run_forever,
            ping_timeout=-3,
            sslopt={"cert_reqs": ssl.CERT_NONE},
        )

    @unittest.skipUnless(TEST_WITH_INTERNET, "Internet-requiring tests are disabled")
    def test_close_status_code(self):
        """Test extraction of close frame status code and close reason in WebSocketApp"""

        def on_close(wsapp, close_status_code, close_msg):
            print("on_close reached")

        app = ws.WebSocketApp(
            "wss://tsock.us1.twilio.com/v3/wsconnect", on_close=on_close
        )
        closeframe = ws.ABNF(
            opcode=ws.ABNF.OPCODE_CLOSE, data=b"\x03\xe8no-init-from-client"
        )
        self.assertEqual([1000, "no-init-from-client"], app._get_close_args(closeframe))

        closeframe = ws.ABNF(opcode=ws.ABNF.OPCODE_CLOSE, data=b"")
        self.assertEqual([None, None], app._get_close_args(closeframe))

        app2 = ws.WebSocketApp("wss://tsock.us1.twilio.com/v3/wsconnect")
        closeframe = ws.ABNF(opcode=ws.ABNF.OPCODE_CLOSE, data=b"")
        self.assertEqual([None, None], app2._get_close_args(closeframe))

        self.assertRaises(
            ws.WebSocketConnectionClosedException,
            app.send,
            data="test if connection is closed",
        )

    def test_parse_close_frame_variants(self):
        """_parse_close_frame decodes bytes, str, and invalid UTF-8 payloads."""

        app = ws.WebSocketApp("ws://example.com")

        binary_frame = ws.ABNF(
            opcode=ws.ABNF.OPCODE_CLOSE, data=b"\x03\xe8normal-closure"
        )
        code, reason = app._parse_close_frame(binary_frame)
        self.assertEqual(code, 1000)
        self.assertEqual(reason, "normal-closure")

        invalid_utf8 = ws.ABNF(
            opcode=ws.ABNF.OPCODE_CLOSE, data=b"\x03\xe8\xff\xfe\xff"
        )
        code, reason = app._parse_close_frame(invalid_utf8)
        self.assertEqual(code, 1000)
        self.assertEqual(reason, "\ufffd\ufffd\ufffd")

        text_data = ws.ABNF(opcode=ws.ABNF.OPCODE_CLOSE, data="ABtext-reason")
        code, reason = app._parse_close_frame(text_data)
        self.assertEqual(code, 0x4142)
        self.assertEqual(reason, "text-reason")

    def test_close_frame_triggers_closed_logic(self):
        """Ensure close frames propagate through closed() to on_error/on_close."""

        close_payload = b"\x03\xe8close-reason"
        close_frame = ws.ABNF(opcode=ws.ABNF.OPCODE_CLOSE, data=close_payload)
        close_results = []
        error_results = []

        def on_close(app, code, reason):
            close_results.append((code, reason))

        def on_error(app, err):
            error_results.append(err)

        class FakeWebSocket:
            def __init__(self, *args, **kwargs):
                self.sock = object()
                self._delivered = False

            def settimeout(self, timeout):
                pass

            def connect(self, *args, **kwargs):
                pass

            def recv_data_frame(self, *args, **kwargs):
                if self._delivered:
                    raise ws.WebSocketConnectionClosedException("closed")
                self._delivered = True
                return (ws.ABNF.OPCODE_CLOSE, close_frame)

            def close(self):
                pass

            def shutdown(self):
                pass

        class FakeDispatcher:
            def __init__(self, app):
                self.app = app

            def read(self, sock, read_callback, check_callback):
                read_callback()

            def reconnect(self, seconds, reconnector):
                pass

        app = ws.WebSocketApp("ws://example.com", on_close=on_close, on_error=on_error)
        fake_dispatcher = FakeDispatcher(app)

        with mock.patch("websocket._app.WebSocket", FakeWebSocket):
            with mock.patch.object(
                ws.WebSocketApp, "create_dispatcher", return_value=fake_dispatcher
            ):
                app.run_forever()

        self.assertEqual(close_results, [(1000, "close-reason")])
        self.assertEqual(len(error_results), 1)
        self.assertIsInstance(error_results[0], ws.WebSocketConnectionClosedException)
        self.assertEqual(getattr(error_results[0], "status_code", None), 1000)
        self.assertEqual(getattr(error_results[0], "reason", None), "close-reason")

    def test_close_exception_legacy_path_has_no_frame(self):
        """Ensure exception path through closed() does not pass a close frame."""

        close_results = []
        error_results = []

        def on_close(app, code, reason):
            close_results.append((code, reason))

        def on_error(app, err):
            error_results.append(err)

        class FakeWebSocket:
            def __init__(self, *args, **kwargs):
                self.sock = mock.Mock()

            def settimeout(self, timeout):
                pass

            def connect(self, *args, **kwargs):
                pass

            def recv_data_frame(self, *args, **kwargs):
                raise ws.WebSocketConnectionClosedException("socket closed")

            def close(self):
                pass

            def shutdown(self):
                pass

        class FakeDispatcher:
            def __init__(self, app):
                self.app = app

            def read(self, sock, read_callback, check_callback):
                read_callback()

            def reconnect(self, seconds, reconnector):
                pass

        app = ws.WebSocketApp("ws://example.com", on_close=on_close, on_error=on_error)
        fake_dispatcher = FakeDispatcher(app)

        with mock.patch("websocket._app.WebSocket", FakeWebSocket):
            with mock.patch.object(
                ws.WebSocketApp, "create_dispatcher", return_value=fake_dispatcher
            ):
                app.run_forever()

        self.assertEqual(close_results, [(None, None)])
        self.assertEqual(len(error_results), 1)
        self.assertIsInstance(error_results[0], ws.WebSocketConnectionClosedException)
        self.assertEqual(str(error_results[0]), "socket closed")
        self.assertFalse(hasattr(error_results[0], "status_code"))

    def test_custom_dispatcher_close_exception(self):
        """Custom dispatchers should route exceptions through closed()."""

        close_results = []
        error_results = []

        def on_close(app, code, reason):
            close_results.append((code, reason))

        def on_error(app, err):
            error_results.append(err)

        class FakeWebSocket:
            def __init__(self, *args, **kwargs):
                self.sock = mock.Mock()

            def settimeout(self, timeout):
                pass

            def connect(self, *args, **kwargs):
                pass

            def recv_data_frame(self, *args, **kwargs):
                raise ws.WebSocketConnectionClosedException("dispatcher closed")

            def close(self):
                pass

            def shutdown(self):
                pass

        class CustomDispatcher:
            def __init__(self):
                self.read_called = False
                self.last_result = None

            def read(self, sock, read_callback, check_callback):
                self.read_called = True
                self.last_result = read_callback()

            def reconnect(self, seconds, reconnector):
                pass

        dispatcher_wrapper = CustomDispatcher()

        app = ws.WebSocketApp("ws://example.com", on_close=on_close, on_error=on_error)

        with mock.patch("websocket._app.WebSocket", FakeWebSocket):
            with mock.patch.object(
                ws.WebSocketApp, "create_dispatcher", return_value=dispatcher_wrapper
            ):
                app.run_forever(dispatcher=object())

        self.assertTrue(dispatcher_wrapper.read_called)
        self.assertTrue(dispatcher_wrapper.last_result)
        self.assertEqual(close_results, [(None, None)])
        self.assertEqual(len(error_results), 1)
        self.assertIsInstance(error_results[0], ws.WebSocketConnectionClosedException)
        self.assertEqual(str(error_results[0]), "dispatcher closed")

    def test_closed_string_argument_converted_to_exception(self):
        """Calling closed() with a string should still reach on_error/on_close."""

        close_results = []
        error_results = []

        def on_close(app, code, reason):
            close_results.append((code, reason))

        def on_error(app, err):
            error_results.append(err)

        class FakeWebSocket:
            def __init__(self, *args, **kwargs):
                self.sock = mock.Mock()

            def settimeout(self, timeout):
                pass

            def connect(self, *args, **kwargs):
                pass

            def shutdown(self):
                pass

            def close(self):
                pass

        class ClosedCallingDispatcher:
            def __init__(self, app, closed_fn):
                self.app = app
                self.closed_fn = closed_fn
                self.read_called = False

            def read(self, sock, read_callback, check_callback):
                self.read_called = True
                self.closed_fn("manual string close")
                self.app.keep_running = False

            def reconnect(self, seconds, reconnector):
                pass

        def fake_create_dispatcher(
            self, ping_timeout, dispatcher, is_ssl, handleDisconnect
        ):
            return ClosedCallingDispatcher(self, handleDisconnect)

        app = ws.WebSocketApp("ws://example.com", on_close=on_close, on_error=on_error)

        with mock.patch("websocket._app.WebSocket", FakeWebSocket):
            with mock.patch.object(
                ws.WebSocketApp, "create_dispatcher", fake_create_dispatcher
            ):
                app.run_forever()

        self.assertEqual(close_results, [(None, None)])
        self.assertEqual(len(error_results), 1)
        self.assertIsInstance(error_results[0], ws.WebSocketConnectionClosedException)
        self.assertEqual(str(error_results[0]), "manual string close")
        self.assertFalse(hasattr(error_results[0], "status_code"))

    @unittest.skipUnless(
        TEST_WITH_LOCAL_SERVER, "Tests using local websocket server are disabled"
    )
    def test_callback_function_exception(self):
        """Test callback function exception handling"""

        exc = None
        passed_app = None

        def on_open(app):
            raise RuntimeError("Callback failed")

        def on_error(app, err):
            nonlocal passed_app
            passed_app = app
            nonlocal exc
            exc = err

        def on_pong(app, _):
            app.close()

        app = ws.WebSocketApp(
            f"ws://127.0.0.1:{LOCAL_WS_SERVER_PORT}",
            on_open=on_open,
            on_error=on_error,
            on_pong=on_pong,
        )
        app.run_forever(ping_interval=2, ping_timeout=1)

        self.assertEqual(passed_app, app)
        self.assertIsInstance(exc, RuntimeError)
        self.assertEqual(str(exc), "Callback failed")

    @unittest.skipUnless(
        TEST_WITH_LOCAL_SERVER, "Tests using local websocket server are disabled"
    )
    def test_callback_method_exception(self):
        """Test callback method exception handling"""

        class Callbacks:
            def __init__(self):
                self.exc = None
                self.passed_app = None
                self.app = ws.WebSocketApp(
                    f"ws://127.0.0.1:{LOCAL_WS_SERVER_PORT}",
                    on_open=self.on_open,
                    on_error=self.on_error,
                    on_pong=self.on_pong,
                )
                self.app.run_forever(ping_interval=2, ping_timeout=1)

            def on_open(self, _):
                raise RuntimeError("Callback failed")

            def on_error(self, app, err):
                self.passed_app = app
                self.exc = err

            def on_pong(self, app, _):
                app.close()

        callbacks = Callbacks()

        self.assertEqual(callbacks.passed_app, callbacks.app)
        self.assertIsInstance(callbacks.exc, RuntimeError)
        self.assertEqual(str(callbacks.exc), "Callback failed")

    @unittest.skipUnless(
        TEST_WITH_LOCAL_SERVER, "Tests using local websocket server are disabled"
    )
    def test_reconnect(self):
        """Test reconnect"""
        pong_count = 0
        exc = None

        def on_error(_, err):
            nonlocal exc
            exc = err

        def on_pong(app, _):
            nonlocal pong_count
            pong_count += 1
            if pong_count == 1:
                # First pong, shutdown socket, enforce read error
                app.sock.shutdown()
            if pong_count >= 2:
                # Got second pong after reconnect
                app.close()

        app = ws.WebSocketApp(
            f"ws://127.0.0.1:{LOCAL_WS_SERVER_PORT}", on_pong=on_pong, on_error=on_error
        )
        app.run_forever(ping_interval=2, ping_timeout=1, reconnect=3)

        self.assertEqual(pong_count, 2)
        self.assertIsInstance(exc, ws.WebSocketTimeoutException)
        self.assertEqual(str(exc), "ping/pong timed out")

    def test_dispatcher_selection_default(self):
        """Test default dispatcher selection"""
        app = ws.WebSocketApp("ws://example.com")

        # Test default dispatcher (non-SSL)
        dispatcher = app.create_dispatcher(ping_timeout=10, is_ssl=False)
        self.assertIsInstance(dispatcher, ws._dispatcher.Dispatcher)

    def test_dispatcher_selection_ssl(self):
        """Test SSL dispatcher selection"""
        app = ws.WebSocketApp("wss://example.com")

        # Test SSL dispatcher
        dispatcher = app.create_dispatcher(ping_timeout=10, is_ssl=True)
        self.assertIsInstance(dispatcher, ws._dispatcher.SSLDispatcher)

    def test_dispatcher_selection_custom(self):
        """Test custom dispatcher selection"""
        from unittest.mock import Mock

        app = ws.WebSocketApp("ws://example.com")
        custom_dispatcher = Mock()
        handle_disconnect = Mock()

        # Test wrapped dispatcher with custom dispatcher
        dispatcher = app.create_dispatcher(
            ping_timeout=10,
            dispatcher=custom_dispatcher,
            handleDisconnect=handle_disconnect,
        )
        self.assertIsInstance(dispatcher, ws._dispatcher.WrappedDispatcher)
        self.assertEqual(dispatcher.dispatcher, custom_dispatcher)
        self.assertEqual(dispatcher.handleDisconnect, handle_disconnect)

    def test_dispatcher_selection_no_ping_timeout(self):
        """Test dispatcher selection without ping timeout"""
        app = ws.WebSocketApp("ws://example.com")

        # Test with None ping_timeout (should default to 10)
        dispatcher = app.create_dispatcher(ping_timeout=None, is_ssl=False)
        self.assertIsInstance(dispatcher, ws._dispatcher.Dispatcher)
        self.assertEqual(dispatcher.ping_timeout, 10)

    def test_suppress_host_parameter(self):
        """Test that suppress_host parameter is accepted by WebSocketApp.run_forever()"""
        app = ws.WebSocketApp("ws://nonexistent.example.com")

        # Test that suppress_host parameter is accepted without error
        # (Connection will fail but parameter should be accepted)
        try:
            app.run_forever(
                suppress_host=True,
                sockopt=((socket.SOL_SOCKET, socket.SO_REUSEADDR, 1),),
            )
        except (ws.WebSocketAddressException, OSError, ConnectionRefusedError):
            # Expected - connection should fail, but parameter was accepted
            pass

        # Test that suppress_host=False also works
        try:
            app.run_forever(
                suppress_host=False,
                sockopt=((socket.SOL_SOCKET, socket.SO_REUSEADDR, 1),),
            )
        except (ws.WebSocketAddressException, OSError, ConnectionRefusedError):
            # Expected - connection should fail, but parameter was accepted
            pass


class WebSocketAppUnitTests(unittest.TestCase):
    def _build_app(self):
        return ws.WebSocketApp("ws://example.com")

    def test_send_helpers_raise_when_socket_inactive(self):
        class DummySock:
            def __init__(self, return_value=0):
                self.return_value = return_value
                self.calls = []

            def send(self, data, opcode):
                self.calls.append((data, opcode))
                return self.return_value

        app = self._build_app()
        app.sock = DummySock(return_value=0)

        with self.assertRaises(ws.WebSocketConnectionClosedException):
            app.send("payload")
        with self.assertRaises(ws.WebSocketConnectionClosedException):
            app.send_text("payload")
        with self.assertRaises(ws.WebSocketConnectionClosedException):
            app.send_bytes(b"payload")

    def test_run_forever_validates_ping_configuration(self):
        app = self._build_app()

        with self.assertRaises(ws.WebSocketException):
            app.run_forever(ping_timeout=0)

        with self.assertRaises(ws.WebSocketException):
            app.run_forever(ping_interval=-1)

        with self.assertRaises(ws.WebSocketException):
            app.run_forever(ping_interval=1, ping_timeout=1)

    def test_run_forever_rejects_when_socket_already_open(self):
        app = self._build_app()
        app.sock = object()

        with self.assertRaises(ws.WebSocketException):
            app.run_forever()

    def test_close_frame_parsing(self):
        app = self._build_app()
        app.on_close = lambda *_: None
        close_payload = struct.pack("!H", 4000) + "bye".encode("utf-8")
        close_frame = ABNF.create_frame(close_payload, ABNF.OPCODE_CLOSE)

        status, reason = app._get_close_args(close_frame)
        self.assertEqual(status, 4000)
        self.assertEqual(reason, "bye")

        empty_frame = ABNF.create_frame(b"", ABNF.OPCODE_CLOSE)
        self.assertEqual(app._get_close_args(empty_frame), [None, None])

    def test_client_initiated_close_captures_peer_close_frame(self):
        """Test that client-initiated close uses stored close frame in teardown."""
        close_results = []

        def on_close(app, code, reason):
            close_results.append((code, reason))

        peer_close_payload = struct.pack("!H", 1000) + "peer-goodbye".encode("utf-8")
        peer_close_frame = ABNF.create_frame(peer_close_payload, ABNF.OPCODE_CLOSE)

        class FakeWebSocket:
            def __init__(self, *args, **kwargs):
                self.sock = mock.Mock()
                self.close_frame = None

            def settimeout(self, timeout):
                pass

            def connect(self, *args, **kwargs):
                pass

            def close(self, **kwargs):
                # Simulate peer sending a close frame in response to client close
                self.close_frame = peer_close_frame

            def shutdown(self):
                pass

            def recv_data_frame(self, *args, **kwargs):
                raise ws.WebSocketConnectionClosedException("closed")

        def fake_read(self, sock, read_callback, check_callback):
            # Trigger a client-initiated close, then let read() call teardown().
            self.app.close()
            read_callback()

        app = ws.WebSocketApp("ws://example.com", on_close=on_close)

        with mock.patch("websocket._app.WebSocket", FakeWebSocket):
            with mock.patch("websocket._app.Dispatcher.read", new=fake_read):
                app.run_forever()

        self.assertEqual(close_results, [(1000, "peer-goodbye")])

    def test_close_captures_frame_when_teardown_clears_socket(self):
        """close() retains its socket while concurrent teardown clears app.sock."""
        app = self._build_app()
        peer_close_frame = ABNF.create_frame(
            struct.pack("!H", 1000) + b"peer-goodbye", ABNF.OPCODE_CLOSE
        )
        close_started = threading.Event()
        allow_close_to_return = threading.Event()
        errors = []

        class FakeWebSocket:
            close_frame = None

            def close(self, **kwargs):
                self.close_frame = peer_close_frame
                close_started.set()
                if not allow_close_to_return.wait(timeout=1):
                    raise TimeoutError("test did not release FakeWebSocket.close()")

        app.sock = FakeWebSocket()

        def close_app():
            try:
                app.close()
            except Exception as error:
                errors.append(error)

        close_thread = threading.Thread(target=close_app)
        close_thread.start()
        self.assertTrue(close_started.wait(timeout=1))

        # This models run_forever() teardown clearing the shared app reference.
        app.sock = None
        allow_close_to_return.set()
        close_thread.join(timeout=1)

        self.assertFalse(close_thread.is_alive())
        self.assertEqual(errors, [])
        self.assertIs(app.last_close_frame, peer_close_frame)

    def test_last_close_frame_reset_on_reconnect(self):
        """Test that last_close_frame is reset when initializing a new socket."""
        app = self._build_app()

        # Simulate a previous close frame being stored
        old_close_payload = struct.pack("!H", 1001) + "old-reason".encode("utf-8")
        app.last_close_frame = ABNF.create_frame(old_close_payload, ABNF.OPCODE_CLOSE)

        # Verify last_close_frame is set
        self.assertIsNotNone(app.last_close_frame)

        # The reset happens in initialize_socket(), which is a nested function
        # inside run_forever(). We verify the attribute exists and can be reset.
        app.last_close_frame = None
        self.assertIsNone(app.last_close_frame)

    def test_callback_errors_propagate_to_on_error_once(self):
        capture = []

        def on_error(app, err):
            capture.append(err)

        def failing_callback(app, *_):
            raise RuntimeError("boom")

        app = self._build_app()
        app.on_error = on_error
        app._callback(failing_callback)
        self.assertEqual(len(capture), 1)
        self.assertIsInstance(capture[0], RuntimeError)

        # When the failing callback is on_error itself, ensure no recursion occurs.
        capture.clear()
        app.on_error = failing_callback
        app._callback(app.on_error)
        self.assertEqual(capture, [])


if __name__ == "__main__":
    unittest.main()



import ipaddress
import os
from typing import Optional, Union
from urllib.parse import unquote, urlparse
from ._exceptions import WebSocketProxyException

"""
_url.py
websocket - WebSocket client library for Python

Copyright 2026 engn33r

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

__all__ = ["parse_url", "get_proxy_info"]


def parse_url(url: str) -> tuple:
    """
    parse url and the result is tuple of
    (hostname, port, resource path and the flag of secure mode)

    Parameters
    ----------
    url: str
        url string.
    """
    if ":" not in url:
        raise ValueError("url is invalid")

    scheme, url = url.split(":", 1)

    parsed = urlparse(url, scheme="http")
    if parsed.hostname:
        hostname = parsed.hostname
    else:
        raise ValueError("hostname is invalid")
    port = 0
    if parsed.port:
        port = parsed.port

    is_secure = False
    if scheme == "ws":
        if not port:
            port = 80
    elif scheme == "wss":
        is_secure = True
        if not port:
            port = 443
    else:
        raise ValueError("scheme %s is invalid" % scheme)

    if parsed.path:
        resource = parsed.path
    else:
        resource = "/"

    if parsed.query:
        resource += f"?{parsed.query}"

    return hostname, port, resource, is_secure


def _is_ip_address(addr: str) -> bool:
    if not isinstance(addr, str):
        raise TypeError("_is_ip_address() argument 1 must be str")
    try:
        ipaddress.ip_address(addr)
    except ValueError:
        return False
    else:
        return True


def _is_subnet_address(hostname: str) -> bool:
    try:
        ipaddress.ip_network(hostname)
    except ValueError:
        return False
    else:
        return True


def _is_address_in_network(ip: str, net: str) -> bool:
    try:
        ip_net: Union[ipaddress.IPv4Network, ipaddress.IPv6Network] = (
            ipaddress.ip_network(ip)
        )
        target_net: Union[ipaddress.IPv4Network, ipaddress.IPv6Network] = (
            ipaddress.ip_network(net)
        )
        return ip_net.subnet_of(target_net)  # type: ignore[arg-type]
    except (TypeError, ValueError):
        return False


def _is_no_proxy_host(hostname: str, no_proxy: Optional[list[str]]) -> bool:
    if not no_proxy:
        if v := os.environ.get("no_proxy", os.environ.get("NO_PROXY", "")).replace(
            " ", ""
        ):
            no_proxy = v.split(",")

    if not no_proxy:
        no_proxy = []

    if "*" in no_proxy:
        return True
    if hostname in no_proxy:
        return True
    if _is_ip_address(hostname):
        return any(
            [
                _is_address_in_network(hostname, subnet)
                for subnet in no_proxy
                if _is_subnet_address(subnet)
            ]
        )
    # Check domain suffix matching - handle both .domain.com and domain.com formats
    # This makes behavior consistent with urllib and other Python libraries
    for domain in no_proxy:
        if domain.startswith("."):
            # Handle .domain.com format
            stripped_domain = domain.lstrip(".")
            if hostname.endswith(stripped_domain):
                return True
        else:
            # Handle domain.com format (should match subdomains too)
            # E.g., "example.com" should match "sub.example.com"
            if hostname == domain or hostname.endswith("." + domain):
                return True
    return False


def get_proxy_info(
    hostname: str,
    is_secure: bool,
    proxy_host: Optional[str] = None,
    proxy_port: int = 0,
    proxy_auth: Optional[tuple] = None,
    no_proxy: Optional[list[str]] = None,
    proxy_type: str = "http",
) -> tuple:
    """
    Try to retrieve proxy host and port from environment
    if not provided in options.
    Result is (proxy_host, proxy_port, proxy_auth).
    proxy_auth is tuple of username and password
    of proxy authentication information.

    Parameters
    ----------
    hostname: str
        Websocket server name.
    is_secure: bool
        Is the connection secure? (wss) looks for "https_proxy" in env
        instead of "http_proxy"
    proxy_host: str
        http proxy host name.
    proxy_port: str or int
        http proxy port. Required when proxy_host is set.
    no_proxy: list
        Whitelisted host names that don't use the proxy.
    proxy_auth: tuple
        HTTP proxy auth information. Tuple of username and password. Default is None.
    proxy_type: str
        Specify the proxy protocol (http, socks4, socks4a, socks5, socks5h). Default is "http".
        Use socks4a or socks5h if you want to send DNS requests through the proxy.
    """
    if _is_no_proxy_host(hostname, no_proxy):
        return None, 0, None

    if proxy_host:
        if not proxy_port:
            raise WebSocketProxyException("Cannot use port 0 when proxy_host specified")
        port = proxy_port
        auth = proxy_auth
        return proxy_host, port, auth

    env_key = "https_proxy" if is_secure else "http_proxy"
    value = os.environ.get(env_key, os.environ.get(env_key.upper(), "")).replace(
        " ", ""
    )
    if value:
        proxy = urlparse(value)
        auth = (
            (unquote(proxy.username or ""), unquote(proxy.password or ""))
            if proxy.username
            else None
        )
        return proxy.hostname, proxy.port, auth

    return None, 0, None



import logging
from typing import Any, Optional

"""
_logging.py
websocket - WebSocket client library for Python

Copyright 2026 engn33r

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

_logger = logging.getLogger("websocket")
try:
    from logging import NullHandler
except ImportError:

    class NullHandler(logging.Handler):  # type: ignore[no-redef]
        def emit(self, record: Any) -> None:
            pass


_logger.addHandler(NullHandler())

_traceEnabled = False
_trace_handler: Optional[logging.Handler] = None
_default_level = _logger.getEffectiveLevel()

__all__ = [
    "enableTrace",
    "dump",
    "error",
    "warning",
    "debug",
    "trace",
    "isEnabledForError",
    "isEnabledForDebug",
    "isEnabledForTrace",
]


def enableTrace(
    traceable: bool,
    handler: Optional[logging.Handler] = None,
    level: str = "DEBUG",
) -> None:
    """
    Turn on/off the traceability.

    Parameters
    ----------
    traceable: bool
        If set to True, traceability is enabled.
    """
    global _traceEnabled, _trace_handler
    if traceable:
        if handler is None:
            handler = logging.StreamHandler()
        if _trace_handler and _trace_handler in _logger.handlers:
            _logger.removeHandler(_trace_handler)
        _trace_handler = handler
        if handler not in _logger.handlers:
            _logger.addHandler(handler)
        _logger.setLevel(getattr(logging, level))
        _traceEnabled = True
    else:
        _traceEnabled = False
        if _trace_handler and _trace_handler in _logger.handlers:
            _logger.removeHandler(_trace_handler)
        _trace_handler = None
        _logger.setLevel(_default_level)


def dump(title: str, message: str) -> None:
    if _traceEnabled:
        _logger.debug(f"--- {title} ---")
        _logger.debug(message)
        _logger.debug("-----------------------")


def error(msg: str) -> None:
    _logger.error(msg)


def warning(msg: str) -> None:
    _logger.warning(msg)


def debug(msg: str) -> None:
    _logger.debug(msg)


def info(msg: str) -> None:
    _logger.info(msg)


def trace(msg: str) -> None:
    if _traceEnabled:
        _logger.debug(msg)


def isEnabledForError() -> bool:
    return _logger.isEnabledFor(logging.ERROR)


def isEnabledForDebug() -> bool:
    return _logger.isEnabledFor(logging.DEBUG)


def isEnabledForTrace() -> bool:
    return _traceEnabled



# -*- coding: utf-8 -*-
#
import os
import os.path
import socket
import ssl
import unittest
from types import SimpleNamespace
from unittest import mock

import websocket
from websocket._exceptions import (
    WebSocketAddressException,
    WebSocketProxyException,
    WebSocketException,
)
from websocket._http import (
    _get_addrinfo_list,
    _start_proxied_socket,
    _tunnel,
    _wrap_sni_socket,
    connect,
    proxy_info,
    read_headers,
    HAVE_PYTHON_SOCKS,
)

"""
test_http.py
websocket - WebSocket client library for Python

Copyright 2026 engn33r

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

try:
    from python_socks._errors import ProxyConnectionError, ProxyError, ProxyTimeoutError
except:
    from websocket._http import ProxyConnectionError, ProxyError, ProxyTimeoutError

# Skip test to access the internet unless TEST_WITH_INTERNET == 1
TEST_WITH_INTERNET = os.environ.get("TEST_WITH_INTERNET", "0") == "1"
TEST_WITH_PROXY = os.environ.get("TEST_WITH_PROXY", "0") == "1"
# Skip tests relying on local websockets server unless LOCAL_WS_SERVER_PORT != -1
LOCAL_WS_SERVER_PORT = os.environ.get("LOCAL_WS_SERVER_PORT", "-1")
TEST_WITH_LOCAL_SERVER = LOCAL_WS_SERVER_PORT != "-1"


class SockMock:
    def __init__(self):
        self.data = []
        self.sent = []

    def add_packet(self, data):
        self.data.append(data)

    def gettimeout(self):
        return None

    def recv(self, bufsize):
        if self.data:
            e = self.data.pop(0)
            if isinstance(e, Exception):
                raise e
            if len(e) > bufsize:
                self.data.insert(0, e[bufsize:])
            return e[:bufsize]

    def send(self, data):
        self.sent.append(data)
        return len(data)

    def close(self):
        pass


class HeaderSockMock(SockMock):
    def __init__(self, fname):
        SockMock.__init__(self)
        path = os.path.join(os.path.dirname(__file__), fname)
        with open(path, "rb") as f:
            self.add_packet(f.read())


class OptsList:
    def __init__(self):
        self.timeout = 1
        self.sockopt = []
        self.sslopt = {"cert_reqs": ssl.CERT_NONE}


class HttpTest(unittest.TestCase):
    def test_read_header(self):
        status, header, _ = read_headers(HeaderSockMock("data/header01.txt"))
        self.assertEqual(status, 101)
        self.assertEqual(header["connection"], "Upgrade")
        # header02.txt is intentionally malformed
        self.assertRaises(
            WebSocketException, read_headers, HeaderSockMock("data/header02.txt")
        )

    def test_read_header_malformed_status_line(self):
        """Malformed responses surface as WebSocketException, not raw
        IndexError, ValueError, or UnicodeDecodeError."""
        one_token = SockMock()
        one_token.add_packet(b"GARBAGE\r\n\r\n")
        with self.assertRaises(WebSocketException) as cm:
            read_headers(one_token)
        self.assertIn("Invalid status line", str(cm.exception))

        bad_status = SockMock()
        bad_status.add_packet(b"HTTP/1.1 XX Switching Protocols\r\n\r\n")
        with self.assertRaises(WebSocketException) as cm:
            read_headers(bad_status)
        self.assertIn("Invalid status line", str(cm.exception))

        bad_utf8 = SockMock()
        bad_utf8.add_packet(b"HTTP/1.1\xff 101 Switching Protocols\r\n\r\n")
        with self.assertRaises(WebSocketException) as cm:
            read_headers(bad_utf8)
        self.assertIn("UTF-8", str(cm.exception))
        self.assertIn("at byte 8", str(cm.exception))

    def test_get_resp_headers_invalid_content_length(self):
        """A non-numeric Content-Length on an error response surfaces as
        WebSocketException, not raw ValueError."""
        from websocket._handshake import _get_resp_headers

        sock = SockMock()
        sock.add_packet(b"HTTP/1.1 400 Bad Request\r\nContent-Length: abc\r\n\r\n")
        with self.assertRaises(WebSocketException) as cm:
            _get_resp_headers(sock)
        self.assertIn("Invalid content-length", str(cm.exception))

    def test_tunnel(self):
        self.assertRaises(
            WebSocketProxyException,
            _tunnel,
            HeaderSockMock("data/header01.txt"),
            "example.com",
            80,
            ("username", "password"),
        )
        self.assertRaises(
            WebSocketProxyException,
            _tunnel,
            HeaderSockMock("data/header02.txt"),
            "example.com",
            80,
            ("username", "password"),
        )

    def test_connect(self):
        # Test proxy error handling without requiring internet - using fast local addresses and short timeouts
        if HAVE_PYTHON_SOCKS:
            # Test SOCKS proxy types with non-routable addresses (fast failure)
            for proxy_type in ["socks4", "socks4a", "socks5", "socks5h"]:
                self.assertRaises(
                    (ProxyTimeoutError, OSError, ConnectionRefusedError),
                    _start_proxied_socket,
                    "wss://127.0.0.1",  # Use localhost instead of example.com
                    OptsList(),
                    proxy_info(
                        http_proxy_host="127.0.0.1",  # Use localhost
                        http_proxy_port="9999",  # Non-existent port for fast failure
                        proxy_type=proxy_type,
                        http_proxy_timeout=0.1,  # Very short timeout
                    ),
                )

            # Test SOCKS connection error with guaranteed-closed port
            self.assertRaises(
                (ProxyConnectionError, ConnectionRefusedError, OSError),
                connect,
                "wss://127.0.0.1",
                OptsList(),
                proxy_info(
                    http_proxy_host="127.0.0.1",
                    http_proxy_port=9999,
                    proxy_type="socks4",
                    http_proxy_timeout=0.1,  # Very short timeout
                ),
                None,
            )

        # Test TypeError with None hostname (no network required)
        self.assertRaises(
            WebSocketAddressException,
            _get_addrinfo_list,
            None,
            80,
            True,
            proxy_info(
                http_proxy_host="127.0.0.1", http_proxy_port="9999", proxy_type="http"
            ),
        )

        # Test HTTP proxy timeout with non-existent port (fast failure)
        self.assertRaises(
            (socket.timeout, ConnectionRefusedError, OSError),
            connect,
            "wss://127.0.0.1",  # Use localhost
            OptsList(),
            proxy_info(
                http_proxy_host="127.0.0.1",  # Use localhost
                http_proxy_port=9999,  # Non-existent port
                proxy_type="http",
                http_proxy_timeout=0.1,  # Very short timeout
            ),
            None,
        )

    @unittest.skipUnless(TEST_WITH_INTERNET, "Internet-requiring tests are disabled")
    def test_connect_with_internet(self):
        # Separate test for cases that actually need internet
        self.assertEqual(
            connect(
                "wss://google.com",
                OptsList(),
                proxy_info(
                    http_proxy_host="8.8.8.8", http_proxy_port=8080, proxy_type="http"
                ),
                True,
            ),
            (True, ("google.com", 443, "/")),
        )

    def test_proxy_connect(self):
        """Proxy logic should exercise SOCKS path and HTTP tunnel path offline."""

        proxy_cfg = proxy_info(
            http_proxy_host="proxy.local",
            http_proxy_port="1080",
            proxy_type="socks5",
            http_proxy_auth=("user", "pass"),
        )
        options = OptsList()
        proxy_instance = mock.Mock()
        proxied_socket = mock.Mock()
        ssl_wrapped_socket = mock.Mock()
        proxy_instance.connect.return_value = proxied_socket

        proxy_types = SimpleNamespace(SOCKS4="socks4", SOCKS5="socks5")

        with mock.patch("websocket._http.HAVE_PYTHON_SOCKS", True), mock.patch.object(
            websocket._http, "ProxyType", proxy_types
        ), mock.patch("websocket._http.Proxy", create=True) as proxy_cls, mock.patch(
            "websocket._http._ssl_socket", return_value=ssl_wrapped_socket
        ) as ssl_socket:
            proxy_cls.create.return_value = proxy_instance
            sock, addr = _start_proxied_socket(
                "wss://example.com/ws", options, proxy_cfg
            )

        proxy_cls.create.assert_called_once()
        proxy_instance.connect.assert_called_once_with("example.com", 443, timeout=None)
        ssl_socket.assert_called_once_with(
            proxied_socket, options.sslopt, "example.com"
        )
        self.assertIs(sock, ssl_wrapped_socket)
        self.assertEqual(addr, ("example.com", 443, "/ws"))

        fake_sock = mock.Mock()
        with mock.patch("websocket._http.send") as send_mock, mock.patch(
            "websocket._http.read_headers", return_value=(200, {}, "OK")
        ):
            returned = _tunnel(fake_sock, "endpoint", 9000, ("demo", "secret"))

        self.assertIs(returned, fake_sock)
        send_mock.assert_called_once()

    def test_proxy_connect_no_port_raises(self):
        """A SOCKS proxy with an explicit host but no port raises the same
        WebSocketProxyException as the http-proxy path, instead of passing
        port 0 (or None) to python-socks."""
        for proxy_cfg in (
            proxy_info(http_proxy_host="proxy.local", proxy_type="socks5"),
            proxy_info(
                http_proxy_host="proxy.local",
                http_proxy_port=None,
                proxy_type="socks5",
            ),
        ):
            with self.assertRaises(WebSocketProxyException):
                _start_proxied_socket("ws://example.com/ws", OptsList(), proxy_cfg)

    @unittest.skipUnless(TEST_WITH_INTERNET, "Internet-requiring tests are disabled")
    def test_sslopt(self):
        ssloptions = {
            "check_hostname": False,
            "server_hostname": "ServerName",
            "ssl_version": ssl.PROTOCOL_TLS_CLIENT,
            "ciphers": "TLS_AES_256_GCM_SHA384:TLS_CHACHA20_POLY1305_SHA256:\
                        TLS_AES_128_GCM_SHA256:ECDHE-ECDSA-AES256-GCM-SHA384:\
                        ECDHE-RSA-AES256-GCM-SHA384:DHE-RSA-AES256-GCM-SHA384:\
                        ECDHE-ECDSA-CHACHA20-POLY1305:ECDHE-RSA-CHACHA20-POLY1305:\
                        DHE-RSA-CHACHA20-POLY1305:ECDHE-ECDSA-AES128-GCM-SHA256:\
                        ECDHE-RSA-AES128-GCM-SHA256:DHE-RSA-AES128-GCM-SHA256:\
                        ECDHE-ECDSA-AES256-SHA384:ECDHE-RSA-AES256-SHA384:\
                        DHE-RSA-AES256-SHA256:ECDHE-ECDSA-AES128-SHA256:\
                        ECDHE-RSA-AES128-SHA256:DHE-RSA-AES128-SHA256:\
                        ECDHE-ECDSA-AES256-SHA:ECDHE-RSA-AES256-SHA",
            "ecdh_curve": "prime256v1",
        }
        ws_ssl1 = websocket.WebSocket(sslopt=ssloptions)
        ws_ssl1.connect("wss://api.bitfinex.com/ws/2")
        ws_ssl1.send("Hello")
        ws_ssl1.close()

        ws_ssl2 = websocket.WebSocket(sslopt={"check_hostname": True})
        ws_ssl2.connect("wss://api.bitfinex.com/ws/2")
        ws_ssl2.close()

    def test_proxy_info(self):
        self.assertEqual(
            proxy_info(
                http_proxy_host="127.0.0.1", http_proxy_port="8080", proxy_type="http"
            ).proxy_protocol,
            "http",
        )
        self.assertRaises(
            ProxyError,
            proxy_info,
            http_proxy_host="127.0.0.1",
            http_proxy_port="8080",
            proxy_type="badval",
        )
        self.assertEqual(
            proxy_info(
                http_proxy_host="example.com", http_proxy_port="8080", proxy_type="http"
            ).proxy_host,
            "example.com",
        )
        self.assertEqual(
            proxy_info(
                http_proxy_host="127.0.0.1", http_proxy_port="8080", proxy_type="http"
            ).proxy_port,
            "8080",
        )
        self.assertEqual(
            proxy_info(
                http_proxy_host="127.0.0.1", http_proxy_port="8080", proxy_type="http"
            ).auth,
            None,
        )
        self.assertEqual(
            proxy_info(
                http_proxy_host="127.0.0.1",
                http_proxy_port="8080",
                proxy_type="http",
                http_proxy_auth=("my_username123", "my_pass321"),
            ).auth[0],
            "my_username123",
        )
        self.assertEqual(
            proxy_info(
                http_proxy_host="127.0.0.1",
                http_proxy_port="8080",
                proxy_type="http",
                http_proxy_auth=("my_username123", "my_pass321"),
            ).auth[1],
            "my_pass321",
        )
        self.assertIsNone(proxy_info().proxy_timeout)


class HttpPureUnitTests(unittest.TestCase):
    def test_get_addrinfo_list_uses_proxy_host(self):
        proxy = proxy_info(
            http_proxy_host="proxy.example", http_proxy_port=8080, proxy_type="http"
        )

        with mock.patch(
            "websocket._http.socket.getaddrinfo", return_value=[("addr",)]
        ) as mocked_getaddrinfo:
            addrinfo, need_tunnel, auth = _get_addrinfo_list(
                "realhost.example", 443, True, proxy
            )

        self.assertTrue(need_tunnel)
        self.assertIsNone(auth)
        mocked_getaddrinfo.assert_called_once_with(
            "proxy.example", 8080, 0, socket.SOCK_STREAM, socket.SOL_TCP
        )
        self.assertEqual(addrinfo, [("addr",)])

    def test_get_addrinfo_list_wraps_gaierror(self):
        proxy = proxy_info()

        with mock.patch(
            "websocket._http.socket.getaddrinfo", side_effect=socket.gaierror("boom")
        ):
            with self.assertRaises(WebSocketAddressException):
                _get_addrinfo_list("example.com", 80, False, proxy)

    def test_wrap_sni_socket_raises_for_missing_ca_file(self):
        dummy_sock = mock.Mock()
        sslopt = {"cert_reqs": ssl.CERT_REQUIRED, "ca_certs": "/does/not/exist"}

        with mock.patch(
            "websocket._http.ssl.SSLContext.load_verify_locations",
            side_effect=FileNotFoundError("missing"),
        ):
            with self.assertRaises(WebSocketException) as ctx:
                _wrap_sni_socket(dummy_sock, sslopt, "example.com", True)

        self.assertIn("SSL CA certificate loading failed", str(ctx.exception))

    def test_start_proxied_socket_requires_python_socks(self):
        proxy = proxy_info(
            http_proxy_host="127.0.0.1", http_proxy_port=1080, proxy_type="socks5"
        )
        options = SimpleNamespace(sslopt={})

        with mock.patch("websocket._http.HAVE_PYTHON_SOCKS", False):
            with self.assertRaises(WebSocketException):
                _start_proxied_socket("wss://example.com", options, proxy)

    def test_tunnel_raises_for_non_200(self):
        fake_socket = object()

        with mock.patch("websocket._http.send") as send_mock, mock.patch(
            "websocket._http.read_headers", return_value=(403, {}, "Forbidden")
        ):
            with self.assertRaises(WebSocketProxyException):
                _tunnel(fake_socket, "example.com", 443, ("user", "pass"))

        send_mock.assert_called()


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
#
import importlib
import sys
import unittest
from unittest import mock

from websocket._abnf import ABNF, continuous_frame, frame_buffer
from websocket._exceptions import WebSocketPayloadException, WebSocketProtocolException

"""
test_abnf.py
websocket - WebSocket client library for Python

Copyright 2026 engn33r

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


class ABNFTest(unittest.TestCase):
    def test_init(self):
        a = ABNF(0, 0, 0, 0, opcode=ABNF.OPCODE_PING)
        self.assertEqual(a.fin, 0)
        self.assertEqual(a.rsv1, 0)
        self.assertEqual(a.rsv2, 0)
        self.assertEqual(a.rsv3, 0)
        self.assertEqual(a.opcode, 9)
        self.assertEqual(a.data, "")
        a_bad = ABNF(0, 1, 0, 0, opcode=77)
        self.assertEqual(a_bad.rsv1, 1)
        self.assertEqual(a_bad.opcode, 77)

    def test_status_constants_exported(self):
        """Every STATUS_ constant in _abnf must be exported via __all__."""
        import websocket
        import websocket._abnf as abnf_mod

        defined = {n for n in dir(abnf_mod) if n.startswith("STATUS_")}
        exported = {n for n in abnf_mod.__all__ if n.startswith("STATUS_")}
        self.assertSetEqual(exported, defined)
        self.assertEqual(websocket.STATUS_SERVICE_RESTART, 1012)
        self.assertEqual(websocket.STATUS_TRY_AGAIN_LATER, 1013)

    def test_validate(self):
        a_invalid_ping = ABNF(0, 0, 0, 0, opcode=ABNF.OPCODE_PING)
        self.assertRaises(
            WebSocketProtocolException,
            a_invalid_ping.validate,
            skip_utf8_validation=False,
        )
        a_bad_rsv_value = ABNF(0, 1, 0, 0, opcode=ABNF.OPCODE_TEXT)
        self.assertRaises(
            WebSocketProtocolException,
            a_bad_rsv_value.validate,
            skip_utf8_validation=False,
        )
        a_bad_opcode = ABNF(0, 0, 0, 0, opcode=77)
        self.assertRaises(
            WebSocketProtocolException,
            a_bad_opcode.validate,
            skip_utf8_validation=False,
        )
        a_bad_close_frame = ABNF(0, 0, 0, 0, opcode=ABNF.OPCODE_CLOSE, data=b"\x01")
        self.assertRaises(
            WebSocketProtocolException,
            a_bad_close_frame.validate,
            skip_utf8_validation=False,
        )
        a_bad_close_frame_2 = ABNF(
            0, 0, 0, 0, opcode=ABNF.OPCODE_CLOSE, data=b"\x01\x8a\xaa\xff\xdd"
        )
        self.assertRaises(
            WebSocketProtocolException,
            a_bad_close_frame_2.validate,
            skip_utf8_validation=False,
        )
        a_bad_close_frame_3 = ABNF(
            0, 0, 0, 0, opcode=ABNF.OPCODE_CLOSE, data=b"\x03\xe7"
        )
        self.assertRaises(
            WebSocketProtocolException,
            a_bad_close_frame_3.validate,
            skip_utf8_validation=True,
        )

    def test_validate_exception_messages_are_formatted(self):
        invalid_opcode = ABNF(0, 0, 0, 0, opcode=5)
        with self.assertRaisesRegex(WebSocketProtocolException, r"^Invalid opcode 5$"):
            invalid_opcode.validate()

        invalid_close_status = ABNF(
            1, 0, 0, 0, opcode=ABNF.OPCODE_CLOSE, mask_value=0, data=b"\x03\xe7"
        )
        with self.assertRaisesRegex(
            WebSocketProtocolException, r"^Invalid close opcode 999$"
        ):
            invalid_close_status.validate()

    def test_mask(self):
        abnf_none_data = ABNF(
            0, 0, 0, 0, opcode=ABNF.OPCODE_PING, mask_value=1, data=None
        )
        bytes_val = b"aaaa"
        self.assertEqual(abnf_none_data._get_masked(bytes_val), bytes_val)
        abnf_str_data = ABNF(
            0, 0, 0, 0, opcode=ABNF.OPCODE_PING, mask_value=1, data="a"
        )
        self.assertEqual(abnf_str_data._get_masked(bytes_val), b"aaaa\x00")

    def test_format(self):
        abnf_bad_rsv_bits = ABNF(2, 0, 0, 0, opcode=ABNF.OPCODE_TEXT)
        self.assertRaises(ValueError, abnf_bad_rsv_bits.format)
        abnf_bad_opcode = ABNF(0, 0, 0, 0, opcode=5)
        self.assertRaises(ValueError, abnf_bad_opcode.format)
        abnf_length_10 = ABNF(0, 0, 0, 0, opcode=ABNF.OPCODE_TEXT, data="abcdefghij")
        self.assertEqual(b"\x01", abnf_length_10.format()[0].to_bytes(1, "big"))
        self.assertEqual(b"\x8a", abnf_length_10.format()[1].to_bytes(1, "big"))
        self.assertEqual("fin=0 opcode=1 data=abcdefghij", abnf_length_10.__str__())
        abnf_length_20 = ABNF(
            0, 0, 0, 0, opcode=ABNF.OPCODE_BINARY, data="abcdefghijabcdefghij"
        )
        self.assertEqual(b"\x02", abnf_length_20.format()[0].to_bytes(1, "big"))
        self.assertEqual(b"\x94", abnf_length_20.format()[1].to_bytes(1, "big"))
        abnf_no_mask = ABNF(
            0, 0, 0, 0, opcode=ABNF.OPCODE_TEXT, mask_value=0, data=b"\x01\x8a\xcc"
        )
        self.assertEqual(b"\x01\x03\x01\x8a\xcc", abnf_no_mask.format())

    def test_frame_buffer(self):
        fb = frame_buffer(0, True)
        self.assertEqual(fb.recv, 0)
        self.assertEqual(fb.skip_utf8_validation, True)
        fb.header = (1, 0, 0, 0, ABNF.OPCODE_TEXT, 1, 10)
        fb.length = 10
        fb.mask_value = b"abcd"
        fb.clear()
        self.assertIsNone(fb.header)
        self.assertIsNone(fb.length)
        self.assertIsNone(fb.mask_value)
        self.assertFalse(fb.has_mask())

    def test_frame_buffer_recv_frame_handles_chunked_reads(self):
        payload = b"Hello"
        frame_bytes = b"\x81" + bytes([len(payload)]) + payload  # Unmasked text frame
        chunks = [frame_bytes[:1], frame_bytes[1:2], frame_bytes[2:4], frame_bytes[4:]]

        def chunked_recv(_):
            if not chunks:
                raise AssertionError("recv called after buffer drained")
            return chunks.pop(0)

        fb = frame_buffer(chunked_recv, skip_utf8_validation=False)
        frame = fb.recv_frame()

        self.assertEqual(frame.fin, 1)
        self.assertEqual(frame.opcode, ABNF.OPCODE_TEXT)
        self.assertEqual(frame.data, payload)
        self.assertEqual(chunks, [])

    def test_recv_strict_preserves_unconsumed_bytes(self):
        remaining = bytearray(b"abcdef")

        def greedy_recv(_):
            if not remaining:
                raise AssertionError("recv called with empty buffer")
            chunk = bytes(remaining)
            remaining.clear()
            return chunk

        fb = frame_buffer(greedy_recv, skip_utf8_validation=True)
        first = fb.recv_strict(4)
        self.assertEqual(first, b"abcd")
        # The extra bytes should stay buffered for the next strict read
        self.assertEqual(fb.recv_buffer, [b"ef"])

        second = fb.recv_strict(2)
        self.assertEqual(second, b"ef")
        self.assertEqual(fb.recv_buffer, [])

    def test_recv_strict_handles_none_and_non_bytes(self):
        calls_none = [b"ab", None, b"cd"]

        def recv_none(_):
            return calls_none.pop(0)

        fb_none = frame_buffer(recv_none, skip_utf8_validation=True)
        self.assertEqual(fb_none.recv_strict(4), b"ab")
        self.assertEqual(fb_none.recv_strict(2), b"cd")
        self.assertEqual(calls_none, [])

        calls_str = [b"xy", "zz", b"pq"]

        def recv_str(_):
            return calls_str.pop(0)

        fb_str = frame_buffer(recv_str, skip_utf8_validation=True)
        self.assertEqual(fb_str.recv_strict(4), b"xy")
        self.assertEqual(fb_str.recv_strict(2), b"pq")
        self.assertEqual(calls_str, [])


class AbnfPureUnitTests(unittest.TestCase):
    def test_mask_without_wsaccel(self):
        import websocket._abnf as abnf_module

        mask_key = b"\x01\x02\x03\x04"
        data = b"\x05\x06\x07\x08\t"
        expected = bytes(mask_key[i % 4] ^ data[i] for i in range(len(data)))

        with mock.patch.dict(sys.modules, {"wsaccel": None, "wsaccel.xormask": None}):
            reloaded = importlib.reload(abnf_module)
            self.assertEqual(reloaded.ABNF.mask(mask_key, data), expected)

        importlib.reload(reloaded)

    def test_format_encodes_large_lengths(self):
        frame_126 = ABNF(
            1, 0, 0, 0, opcode=ABNF.OPCODE_TEXT, mask_value=0, data=b"x" * 126
        )
        formatted = frame_126.format()
        self.assertEqual(formatted[1], 0x7E)
        self.assertEqual(formatted[2:4], (126).to_bytes(2, "big"))

        frame_65536 = ABNF(
            1, 0, 0, 0, opcode=ABNF.OPCODE_BINARY, mask_value=0, data=b"y" * 66000
        )
        formatted_long = frame_65536.format()
        self.assertEqual(formatted_long[1], 0x7F)
        self.assertEqual(len(formatted_long[2:10]), 8)
        self.assertEqual(int.from_bytes(formatted_long[2:10], "big"), 66000)

    def test_continuous_frame_invalid_utf8_raises(self):
        cont = continuous_frame(fire_cont_frame=False, skip_utf8_validation=False)
        start = ABNF(
            fin=0,
            opcode=ABNF.OPCODE_TEXT,
            data=b"\xf0\x28",
            mask_value=0,
        )
        cont.validate(start)
        cont.add(start)

        end = ABNF(
            fin=1,
            opcode=ABNF.OPCODE_CONT,
            data=b"\x8c\x28",
            mask_value=0,
        )
        cont.validate(end)
        cont.add(end)
        self.assertTrue(cont.is_fire(end))
        with self.assertRaises(WebSocketPayloadException):
            cont.extract(end)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
import unittest
import socket
import threading
import time
from unittest.mock import Mock, patch

from websocket import WebSocket, WebSocketApp
from websocket._core import WebSocket as CoreWebSocket
from websocket._exceptions import WebSocketConnectionClosedException


class ReconnectBadFileDescriptorTest(unittest.TestCase):
    """Test edge cases around reconnection when socket file descriptors become invalid"""

    def test_shutdown_with_already_closed_socket(self):
        """Test that shutdown() handles already closed sockets gracefully (issue #1026)"""
        ws = CoreWebSocket()

        # Create a mock socket
        mock_sock = Mock()
        mock_sock._closed = False
        ws.sock = mock_sock
        ws.connected = True

        # Simulate the socket being closed externally (e.g., network failure)
        mock_sock.close.side_effect = OSError("Bad file descriptor")

        # shutdown() should not raise an exception
        ws.shutdown()

        # Verify cleanup happened
        self.assertIsNone(ws.sock)
        self.assertFalse(ws.connected)

    def test_shutdown_with_closed_socket_attribute(self):
        """Test shutdown() when socket's _closed attribute is True"""
        ws = CoreWebSocket()

        # Create a mock socket that's already closed
        mock_sock = Mock()
        mock_sock._closed = True
        ws.sock = mock_sock
        ws.connected = True

        # shutdown() should not call close() on already closed socket
        ws.shutdown()

        # Verify close() was not called and cleanup happened
        mock_sock.close.assert_not_called()
        self.assertIsNone(ws.sock)
        self.assertFalse(ws.connected)

    def test_shutdown_with_missing_closed_attribute(self):
        """Test shutdown() when socket doesn't have _closed attribute"""
        ws = CoreWebSocket()

        # Create a mock socket without _closed attribute
        mock_sock = Mock()
        del mock_sock._closed  # Remove _closed attribute
        ws.sock = mock_sock
        ws.connected = True

        # shutdown() should handle AttributeError gracefully
        ws.shutdown()

        # Verify cleanup happened despite AttributeError
        self.assertIsNone(ws.sock)
        self.assertFalse(ws.connected)

    def test_shutdown_with_none_socket(self):
        """Test shutdown() when socket is None"""
        ws = CoreWebSocket()
        ws.sock = None
        ws.connected = False

        # Should not raise any exception
        ws.shutdown()

        self.assertIsNone(ws.sock)
        self.assertFalse(ws.connected)

    def test_websocketapp_reconnect_with_bad_fd_simulation(self):
        """Test WebSocketApp reconnection when socket becomes invalid during reconnect"""

        reconnect_attempts = []
        error_messages = []
        connection_count = 0

        def on_error(ws, error):
            error_messages.append(str(error))

        def on_reconnect(ws):
            reconnect_attempts.append(True)

        def on_open(ws):
            nonlocal connection_count
            connection_count += 1

        app = WebSocketApp(
            "ws://echo.websocket.events/",
            on_error=on_error,
            on_reconnect=on_reconnect,
            on_open=on_open,
        )

        # Test the actual shutdown method with bad file descriptor
        with patch("websocket._core.WebSocket.connect") as mock_connect:
            with patch("websocket._core.WebSocket.recv_data_frame") as mock_recv:
                # Simulate successful connection
                mock_connect.return_value = None

                # Simulate connection loss after a brief connection
                mock_recv.side_effect = WebSocketConnectionClosedException(
                    "Connection lost"
                )

                # Create a WebSocket instance with a mock socket that fails on close
                real_ws = CoreWebSocket()
                mock_socket = Mock()
                mock_socket._closed = False
                mock_socket.close.side_effect = OSError("[Errno 9] Bad file descriptor")
                real_ws.sock = mock_socket
                real_ws.connected = True

                # Test that our fixed shutdown method handles this gracefully
                # This should not raise an exception
                real_ws.shutdown()

                # Verify cleanup happened despite the error
                self.assertIsNone(real_ws.sock)
                self.assertFalse(real_ws.connected)

        # Test passes if shutdown() handled bad file descriptor gracefully

    def test_reconnection_with_externally_closed_socket(self):
        """Test the specific scenario from issue #1026: socket killed externally during reconnection"""

        # Create a WebSocket instance
        ws = CoreWebSocket()

        # Create a real socket and then close it externally to simulate network failure
        real_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        ws.sock = real_socket
        ws.connected = True

        # Close the socket externally (simulates `ss --kill` or network failure)
        real_socket.close()

        # Now try to shutdown the WebSocket - this should not raise "Bad file descriptor"
        # Our fix should handle this gracefully
        ws.shutdown()

        # Verify cleanup happened
        self.assertIsNone(ws.sock)
        self.assertFalse(ws.connected)

    def _safe_shutdown(self, ws_instance, mock_socket):
        """Helper method that simulates the fixed shutdown behavior"""
        try:
            if not mock_socket._closed:
                mock_socket.close()
        except (OSError, AttributeError):
            pass
        finally:
            ws_instance.sock = None
            ws_instance.connected = False

    def test_multiple_shutdown_calls(self):
        """Test that multiple shutdown() calls don't cause issues"""
        ws = CoreWebSocket()

        # Create a real socket for more realistic testing
        real_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        ws.sock = real_sock
        ws.connected = True

        # First shutdown should work normally
        ws.shutdown()
        self.assertIsNone(ws.sock)
        self.assertFalse(ws.connected)

        # Second shutdown should not raise exception
        ws.shutdown()
        self.assertIsNone(ws.sock)
        self.assertFalse(ws.connected)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
#
import importlib
import io
import logging
import unittest

import websocket
import websocket._logging as logging_module


class LoggingHelperTests(unittest.TestCase):
    def setUp(self):
        # Reload to ensure a clean logger for each test
        self.logging_mod = importlib.reload(logging_module)
        self.logger = logging.getLogger("websocket")

    def tearDown(self):
        importlib.reload(logging_module)

    def test_enable_trace_toggles_handler_and_level(self):
        stream = io.StringIO()
        handler = logging.StreamHandler(stream)
        original_level = self.logger.level

        self.logging_mod.enableTrace(True, handler=handler)
        self.assertTrue(self.logging_mod.isEnabledForTrace())
        self.assertIn(handler, self.logger.handlers)
        self.assertEqual(self.logger.level, logging.DEBUG)

        self.logging_mod.enableTrace(False)
        self.assertFalse(self.logging_mod.isEnabledForTrace())
        self.assertNotIn(handler, self.logger.handlers)
        self.assertEqual(self.logger.level, original_level)

    def test_dump_and_trace_emit_only_when_enabled(self):
        stream = io.StringIO()
        handler = logging.StreamHandler(stream)

        self.logging_mod.enableTrace(True, handler=handler)
        self.logging_mod.dump("header", "payload")
        self.logging_mod.trace("trace-line")
        handler.flush()
        logged = stream.getvalue()
        self.assertIn("header", logged)
        self.assertIn("payload", logged)
        self.assertIn("trace-line", logged)

        stream.truncate(0)
        stream.seek(0)
        self.logging_mod.enableTrace(False)
        self.logging_mod.dump("hidden", "secret")
        self.logging_mod.trace("should-not-log")
        handler.flush()
        self.assertEqual(stream.getvalue(), "")


class WebSocketModuleTraceTests(unittest.TestCase):
    def tearDown(self):
        websocket.enableTrace(False)

    def test_public_enable_trace_controls_logging(self):
        stream = io.StringIO()
        handler = logging.StreamHandler(stream)

        websocket.enableTrace(True, handler=handler)
        websocket.dump("public", "message")
        websocket.trace("visible-trace")
        handler.flush()
        logged = stream.getvalue()
        self.assertIn("public", logged)
        self.assertIn("message", logged)
        self.assertIn("visible-trace", logged)

        stream.truncate(0)
        stream.seek(0)
        websocket.enableTrace(False)
        websocket.dump("hidden", "payload")
        websocket.trace("not-logged")
        handler.flush()
        self.assertEqual(stream.getvalue(), "")


if __name__ == "__main__":
    unittest.main()
